
#include <asm/types.h>
#include <asm/io.h>

/*addr is from 0x00 to 0x3F if 64B memory (to 0x7F if 128B) */
#define write_cmos(addr,value)	{ outb_p(addr, 0x70); outb(value, 0x71); }
#define read_cmos(addr)		( outb_p(addr, 0x70), inb(0x71) )

/* test if not updating the clock right now */
/* not using it but might come in handy one day :) */
#define CMOS_SAFERTC	for (;;) {				\
				outb(0x0A, 0x70);		\
				if (inb(0x71)&0x80)		\
					break;	/* ready */	\
			}

#define CMOS_SEC	0
#define CMOS_MIN	2
#define CMOS_HOUR	4
#define CMOS_WEEKDAY	6
#define CMOS_DAY	7
#define CMOS_MONTH	8
#define CMOS_YEAR	9
#define CMOS_CENTURY	0x32

#define CMOS_ALARMSEC	1
#define CMOS_ALARMMIN	3
#define CMOS_ALARMHOUR	5

/* convert from our beloved BCD :) */
#define BCD2NUM(b)	((((b)>>4)*10)+((b)&0xf))

/*
  0Ah         Status Register A
  0Bh         Status Register B
  0Ch         Status Register C
  0Dh         Status Register D
  0Eh         Diagnostic Status Byte
  0Fh         Shutdown Status Byte
  10h         Disk Drive Type for Drives A: and B:
              The drive-type bytes use bits 0:3 for the first
              drive and 4:7 for the other disk drive types.
  00h         no drive present
  01h         double sided 360k
  02h         high capacity (1.2 meg)
  03h-0Fh     reserved
  11h         (AT):Reserved    (PS/2):drive type for hard disk C:
  12h         (PS/2):drive type for hard disk D:
              (AT, XT/286):hard disk type for drives C: and D:
	  Format of drive-type entry for AT, XT/286:
	  0       number of cyls in drive (0-1023 allowed)
	  2       number of heads per drive (0-15 allowed)
	  3       starting reduced write compensation (not used on AT)
	  5       starting cylinder for write compensation
	  7       max. ECC data burst length, XT only
	  8       control byte
	  Bit
	  7       disable disk-access retries
	  6       disable ECC retries
	  5-4     reserved, set to zero
	  3       more than 8 heads
	  2-0     drive option on XT (not used by AT)
	  9       timeout value for XT (not used by AT)
	  12      landing zone cylinder number
	  14      number of sectors per track (default 17, 0-17 allowed)

  13h         Reserved
  14h         Equipment Byte (corresponds to sw. 1 on PC and XT)
  15h-16h     Base Memory Size      (low,high)
  17h-18h     Expansion Memory Size (low,high)
  19h-20h     Reserved
          (PS/2) POS information Model 50 (60 and 80 use a 2k
          CMOS RAM that is not accessible through software)
  21h-2Dh     Reserved (not checksumed)
  2Eh-2Fh     Checksum of Bytes 10 Through 20  (low,high)
  30h-31h     Exp. Memory Size as Det. by POST (low,high)
  32h         Date Century Byte
  33h         Information Flags (set during power-on)
  34h-3Fh     Reserved - Put Your Name Here.
*/

/*should test for bit 7? if yes, convert to 24hr format*/
/*like this:
	if (tmp&0x80)
		return (BCD2NUM(tmp&0x7f)+12) % 24;
	else
		return BCD2NUM(tmp);
*/
inline u32 cmos_hour(void)
{
	u32 tmp = (u32) read_cmos(CMOS_HOUR);
	return BCD2NUM(tmp);
}

inline u32 cmos_min(void)
{
	u32 tmp = (u32) read_cmos(CMOS_MIN);
	return BCD2NUM(tmp);
}

inline u32 cmos_sec(void)
{
	u32 tmp = (u32) read_cmos(CMOS_SEC);
	return BCD2NUM(tmp);
}

inline u32 cmos_year(void)
{
	u32 tmp, year;
	tmp = (u32) read_cmos(CMOS_YEAR);
	year = BCD2NUM(tmp);
	tmp = (u32) read_cmos(CMOS_CENTURY);
	year += BCD2NUM(tmp) * 100;
	
	return year;
}

/*******************************************************/
void testcmos()
{
	char wday[7][4] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

	/*real and correct values*/
	printk("%02x:%02x:%02x\n%s %02x.%02x.%x%02x\n",
		read_cmos(CMOS_HOUR),read_cmos(CMOS_MIN),read_cmos(CMOS_SEC),
		wday[read_cmos(CMOS_WEEKDAY)-1],
		read_cmos(CMOS_DAY),read_cmos(CMOS_MONTH),
		read_cmos(CMOS_CENTURY),read_cmos(CMOS_YEAR));
/*
	for (int i=ticks;;) {
		if (i>ticks)
			continue;
		i+=100;
		printk("%d:%d:%d\n",cmos_hour(),cmos_min(),cmos_sec());
	}
*/

//write_cmos(CMOS_ALARMHOUR,0x20);
//write_cmos(CMOS_ALARMMIN,0x56);
//write_cmos(CMOS_ALARMSEC,0x00);
//u8 tmp=read_cmos(0x0B);
//write_cmos(0x0B,tmp|0x20);	/*alarm on 0x20, periodic 0x40*/

//u8 tmp=read_cmos(0x0B);
//write_cmos(0x0B,tmp|0x40);	/*periodic interrupt enable*/
//write_cmos(0x0A,0x0f|0x70);	/*2Hz, DIV_CTL -- not tested*/

u8 tmp=read_cmos(0x0B);
write_cmos(0x0B,tmp|0x20);	/*alarm int*/
write_cmos(CMOS_ALARMHOUR,0xff);
write_cmos(CMOS_ALARMMIN,0xff);
write_cmos(CMOS_ALARMSEC,0x00);		/*every min at 0 sec*/

read_cmos(0x0C);	/* clears register C (make sure int will appear) */

//printk("\n%d\n",read_cmos(0x0B)&0x02);	/*24hour? 0x02 if yes*/

printk("%x:%x:%x\n",read_cmos(CMOS_ALARMHOUR),read_cmos(CMOS_ALARMMIN),read_cmos(CMOS_ALARMSEC));
/**/

}
