/*
9.8.14  Interrupt 14  Page Fault
This exception occurs when paging is enabled (PG=1) and the processor
detects one of the following conditions while translating a linear address
to a physical address:

    The page-directory or page-table entry needed for the address
     translation has zero in its present bit.

    The current procedure does not have sufficient privilege to access the
     indicated page.

The processor makes available to the page fault handler two items of
information that aid in diagnosing the exception and recovering from it:

    An error code on the stack. The error code for a page fault has a
     format different from that for other exceptions (see Figure 9-8). The
     error code tells the exception handler three things:

     1.  Whether the exception was due to a not present page or to an access
         rights violation.

     2.  Whether the processor was executing at user or supervisor level at
         the time of the exception.

     3.  Whether the memory access that caused the exception was a read or
         write.

    CR2 (control register two). The processor stores in CR2 the linear
     address used in the access that caused the exception (see Figure 9-9).
     The exception handler can use this address to locate the corresponding
     page directory and page table entries. If another page fault can occur
     during execution of the page fault handler, the handler should push CR2
     onto the stack.


Figure 9-8.  Page-Fault Error Code Format

 ͻ
 FieldValue                         Description                        
 Ķ
  U/S   0   The access causing the fault originated when the processor 
            was executing in supervisor mode.                          
                                                                       
        1   The access causing the fault originated when the processor 
            was executing in user mode.                                
                                                                       
  W/R   0   The access causing the fault was a read.                   
                                                                       
        1   The access causing the fault was a write.                  
                                                                       
  P     0   The fault was caused by a not-present page.                
                                                                       
        1   The fault was caused by a page-level protection violation. 
 ͼ

       31                               15             7        3 2 1 0
      ͻ
      UW 
      UNDEFINED//P
      SR 
      ͼ


9.8.14.1  Page Fault During Task Switch

The processor may access any of four segments during a task switch:

  1.  Writes the state of the original task in the TSS of that task.

  2.  Reads the GDT to locate the TSS descriptor of the new task.

  3.  Reads the TSS of the new task to check the types of segment
      descriptors from the TSS.

  4.  May read the LDT of the new task in order to verify the segment
      registers stored in the new TSS.

A page fault can result from accessing any of these segments. In the latter
two cases the exception occurs in the context of the new task. The
instruction pointer refers to the next instruction of the new task, not to
the instruction that caused the task switch. If the design of the operating
system permits page faults to occur during task-switches, the page-fault
handler should be invoked via a task gate.


Figure 9-9.  CR2 Format

      31               23               15               7              0
     ͻ
                                                                        
                           PAGE FAULT LINEAR ADDRESS                    
                                                                        
     ͼ


9.8.14.2  Page Fault with Inconsistent Stack Pointer

Special care should be taken to ensure that a page fault does not cause the
processor to use an invalid stack pointer (SS:ESP). Software written for
earlier processors in the 8086 family often uses a pair of instructions to
change to a new stack; for example:

MOV SS, AX
MOV SP, StackTop

With the 80386, because the second instruction accesses memory, it is
possible to get a page fault after SS has been changed but before SP has
received the corresponding change. At this point, the two parts of the stack
pointer SS:SP (or, for 32-bit programs, SS:ESP) are inconsistent.

The processor does not use the inconsistent stack pointer if the handling
of the page fault causes a stack switch to a well defined stack (i.e., the
handler is a task or a more privileged procedure). However, if the page
fault handler is invoked by a trap or interrupt gate and the page fault
occurs at the same privilege level as the page fault handler, the processor
will attempt to use the stack indicated by the current (invalid) stack
pointer.

In systems that implement paging and that handle page faults within the
faulting task (with trap or interrupt gates), software that executes at the
same privilege level as the page fault handler should initialize a new stack
by using the new LSS instruction rather than an instruction pair shown
above. When the page fault handler executes at privilege level zero (the
normal case), the scope of the problem is limited to privilege-level zero
code, typically the kernel of the operating system.
*/

/*
9.9  Exception Summary


Table 9-6 summarizes the exceptions recognized by the 386.

Table 9-6. Exception Summary


Description               Interrupt   Return Address  Exception     Function That Can Generate
                          Number      Points to       Type          the Exception
                                      Faulting
                                      Instruction

Divide error               0          YES             FAULT         DIV, IDIV
Debug exceptions           1          
Some debug exceptions are traps and some are faults.  The exception
handler can determine which has occurred by examining DR6.  (Refer to
Chapter 12.)               
Some debug exceptions are traps and some are faults.  The exception
handler can determine which has occurred by examining DR6.  (Refer to
Chapter 12.)             Any instruction
Breakpoint                 3          NO              TRAP          One-byte INT 3
Overflow                   4          NO              TRAP          INTO
Bounds check               5          YES             FAULT         BOUND
Invalid opcode             6          YES             FAULT         Any illegal instruction
Coprocessor not available  7          YES             FAULT         ESC, WAIT
Double fault               8          YES             ABORT         Any instruction that can
                                                                    generate an exception
Coprocessor Segment
Overrun                    9          NO              ABORT         Any operand of an ESC
                                                                    instruction that wraps around
                                                                    the end of a segment.
Invalid TSS               10          YES             FAULT
An invalid-TSS fault is not restartable if it occurs during the
processing of an external interrupt.        JMP, CALL, IRET, any interrupt
Segment not present       11          YES             FAULT         Any segment-register modifier
Stack exception           12          YES             FAULT         Any memory reference thru SS
General Protection        13          YES             FAULT/ABORT
All GP faults are restartable. If the fault occurs while attempting to
vector to the handler for an external interrupt, the interrupted program is
restartable, but the interrupt may be lost.  Any memory reference or code
                                                                    fetch
Page fault                14          YES             FAULT         Any memory reference or code
                                                                    fetch
Coprocessor error         16          YES             FAULT
Coprocessor errors are reported as a fault on the first ESC or WAIT
instruction executed after the ESC instruction that caused the error.        ESC, WAIT
Two-byte SW Interrupt     0-255       NO              TRAP          INT n


9.10  Error Code Summary

Table 9-7 summarizes the error information that is available with each
exception.


Table 9-7. Error-Code Summary

Description                       Interrupt     Error Code
                                  Number

Divide error                       0            No
Debug exceptions                   1            No
Breakpoint                         3            No
Overflow                           4            No
Bounds check                       5            No
Invalid opcode                     6            No
Coprocessor not available          7            No
System error                       8            Yes (always 0)
Coprocessor Segment Overrun        9            No
Invalid TSS                       10            Yes
Segment not present               11            Yes
Stack exception                   12            Yes
General protection fault          13            Yes
Page fault                        14            Yes
Coprocessor error                 16            No
Two-byte SW interrupt             0-255         No
*/

#include <asm/types.h>
#include <asm/printk.h>
#include <kernel.h>

char exception_name[][30] = {
	/*ID:	Description:			(double fault) Class: */
	/* 0*/	"Divide error",			/* Contributory Exception */
	/* 1*/	"Debug exception",		/* Benign Exceptions */
	/* 2*/	"NMI",
	/* 3*/	"Breakpoint",
	/* 4*/	"Overflow",
	/* 5*/	"Bounds check",
	/* 6*/	"Invalid opcode",
	/* 7*/	"Coprocessor not available",
	/* 8*/	"Double fault",
	/* 9*/	"Coprocessor Segment Overrun",	/* Contributory Exceptions */
	/*10*/	"Invalid TSS",
	/*11*/	"Segment not present",
	/*12*/	"Stack exception",
	/*13*/	"General protection fault",
	/*14*/	"Page fault",			/* Page Fault */
	/*15*/	"(intel reserved exception)",	/* undocumented */
	/*16*/	"Coprocessor error",		/* Benign Exception */
	/*17*/	"Alignment Check Exception",	/* 486+ */
	/*18*/	"Machine Check Exception",	/* Pentium Pro+ */
	/*19*/	"SIMD Floating Point Exception"	/* Pentium 3+ */
	/*20-31 "Intel reserved, do not use" */
	/*32-	"User defined (non-reserved) interrupts"
		not the architecture interrupts and exceptions */
};

/*abort if DIV err, cs:eip points to THE div instruction*/
/*assign to each exception an POSIX signal*/

typedef struct regdump {
	u32 cr2;	/* so we're not loosing it if another page fault appears */
	u32 eflags;
	u32 ds, es, fs, gs, ss;
	u32 edi, esi, ebp, esp, ebx, edx, ecx, eax;
} regdump_t;

void dump_regs(struct regdump *r)
{
	printk("\teax=%08X\tebx=%08X\tecx=%08X\tedx=%08X\n"
	       "\tesi=%08X\tedi=%08X\tebp=%08X\n"
	       "\t ss=%04X\tesp=%08X\n"
	       "\t ds=%04X\t es=%04X\t fs=%04X\t gs=%04X\n"
	       "     eflags=%08X\n",
	       r->eax, r->ebx, r->ecx, r->edx,
	       r->esi, r->edi, r->ebp,
	       r->ss , r->esp,		/* after privilege transfer */
	       r->ds, r->es, r->fs, r->gs,
	       r->eflags);

	u32 cr0, cr3;
	__asm__ __volatile__("\
		movl %%cr0,%0	;\
		movl %%cr3,%1"
		: "=q" (cr0), "=q" (cr3));
	printk("\tcr0=%08x\tcr2=%08x\tcr3=%08x\n", cr0, r->cr2, cr3);
}

/* our versatile exception handler :) */
void exception(u32 esp3, u32 ss3, u32 ec, u32 eip, u32 cs, u32 e, u32 r)
{
printk("cs=%04x eip=%08x error code=%x\n", cs, eip, ec);
if (ss3)
	printk("STACK FROM BEFORE TRANSFER:\nss:%04x esp:%08x\n", ss3, esp3);
/////
	if (e > 19) {
		printk("Exception 0x%02X: Unknown exception (reserved)\n", e);
		dump_regs((struct regdump *) r);
		/* what now? */
		return;
	}

	printk("Exception 0x%02X: %s\n", e, exception_name[e]);

	dump_regs((struct regdump *) r);
	
u32 db6;
	switch (e) {
		case 0:	/* division error */
			/*///TODO///*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			/*CS:EIP at instruction*/
			/*send SIGFPE signal to application (or die if kernel)*/
			break;
		case 1:	/* debug exception */
/***/			__asm__ __volatile__("movl %%db6, %0" : "=q" (db6));
/***/			printk("db6=%08x\n", db6);
			printk("no further handling implemented. exception not fatal, returning...\n");
			/*warning, could be fault... then abort app ///TODO///*/
			/*CS:EIP at instruction if FAULT, to next if TRAP*/
			break;
		case 2: /* NMI */
			printk("no further handling implemented. exception not fatal, returning...\n");
			break;
		case 3: /* breakpoint */
			printk("no further handling implemented. exception not fatal, returning...\n");
			break;
		case 4: /* overflow */
			printk("no further handling implemented. exception not fatal, returning...\n");
			break;
		case 5: /* BOUND range exceeded */
			/*///TODO///*/
			/*at CS:EIP*/
			/*abort app*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 6: /* invalid opcode */
			/*///TODO///*/
			/*at CS:EIP*/
			/*abort app*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 7: /* no coprocessor */
			/*///TODO///*/
			/*tu sa da spravit emulacia koprocesoru a
			ukladanie stavu koprocesoru - aspon toto spravit*/
			/*skontrolovat ci je TS (a MP moze tiez) v cr0
			nastaveny, ulozit stav, clear TS a pokracovat,
			ale neviem kedy sa nastavuje naspat... asi to
			treba zariadit v TSS
			HIGH PRIORITY TODO*/
			/*at CS:EIP*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 8: /* double fault */
/*has an ERROR CODE of zero */
			/*///TODO///*/
			/* is when 2xPAGE FAULT, 2xCONTRIBUTORY, PF+CONTRIB */
			/*may we continue if it's user mode fault? i think so*/
			/*no CS:EIP provided - ABORT class exception*/
			printk("it's fault, not returning\n");
			/*have to abort the program and change task*/
			cli();
			for (;;)
				;
			break;
		case 9: /* coprocessor segment overrun */
		        /* 387 coprocessor only, other generate int 13 */
			/*ABORT class*/
			/*at CS:EIP, we have to reinit FPU with fninit and abort app*/
			printk("it's fault, not returning\n");
			/*have to abort the program and change task*/
			cli();
			for (;;)
				;
			break;
		case 10: /* invalid TSS */
/*segment selector was pushed on stack + check the EXT flag*/
			/*at CS:EIP - see docs whats there*/
		/*dunno how to deal with this*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 11: /* segment not present */
/*segment selector pushed, EXT flag, IDT flag for non-present int handler*/
			/*shouldn't happen in this kernel,
			abort the app*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 12: /* stack fault */
/*segment selector pushed if non-present segment, 0 if segment limit exceeded*/
			/*shouldn't happen in this kernel,
			we're not using segmentation, abort app*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 13: /* general protection fault */
/*error code pushed*/
		/*tu to je velmi zaujimave, pozriet dokumentaciu!!!
		zda sa ze softwareove interrupty musia mat nastaveny DPL na 3*/
		/*riesenie je, zda sa, jednoduche, abort app*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
/*INY HANDLER?*/
		case 14: /* page fault */
		/*handler should not rely on seg.sel's - i'm pushing on stack!*/
		/*ale ved to by sa mi mali natiahnut kernel veci, nie?!*/
	/*takze aj page fault aj double fault by mal byt TASK gate?
	  ak by nastal page fault a handler by vyvolal double fault
	  tak len tak by sme to zvladli (stacky zvlast v kerneli, ktore
	  sa neswapuju) 
	  ajaj */

/*error code on stack*/
printk("user? %d - write? %d - present? %d\n",(ec>>2)&1,(ec>>1)&1,ec&1);
/*v r->cr2 je linear address faultu co prave vybavujeme*/
		/*porozmyslat o nechani zapnutych preruseni pocas tohoto a mozno inych*/
		/*tu osetrit stack overflow programov (a mozno kernelu),
		swapping, copy-on-write, naloadovanie nejakych stranok a tak*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 16: /* coprocessor error */
		/*nezaujimave*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 17: /* alignment check */
		/*error code 0*/
		/*nezaujimave, asi abort*/
			printk("it's fault, not returning\n");
			cli();
			for (;;)
				;
			break;
		case 18: /* machine check - must be enabled */
		case 19: /* SIMD exception - must be enabled */
			printk("this is not happening right now...:)\nHalt");
			cli();
			for (;;)
				;
			break;
	}

}
